/* Copyright (C) 1999 Chris Vine, G3XXF

This program is distributed under the General Public Licence, version 2.
For particulars of this and relevant disclaimers see the file
COPYRIGHT distributed with the source files.

*/

// main.cpp

#include <unistd.h>
#include <stdlib.h>
#include <iostream.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/time.h>
#include <limits.h>
#include <fstream.h>
#include <string.h>
#include <ctype.h>
#include <ncurses.h>
// we don't want to use the 'clear' macro in ncurses.h - it will interfere with iostream::clear()
#undef clear
#include "qqueue.cpp"   // contains template definition
#include "buffers.h"
#include "com.h"
#include "keyboard.h"
#include "prog_defs.h"
#include "pipes.h"
#include "tnc.h"
#include "screen.h"

#define WORDWRAP 70

///////////////////////// function declarations //////////////////////

void configure_program(int&, int&);
void parent_process(pid_t, Pipe_fifo&, Pipe_fifo&);
void child_process(int, int, Pipe_fifo&, Pipe_fifo&);
void send_parms(WinList&);
void gpl_query(void);
void child_exit_handler(int);

///////////////////////////// global data /////////////////////////////

ifstream filein;
ofstream fileout;
Prog_func prog_func = {Prog_func::word, Prog_func::normal, FALSE, 0, 0, 0, {0}, FALSE, FALSE};
int child_exit_flag;

//////////////////////////////////////////////////////////////////////

main(int argc, char* argv[]) {

    int port;
    int speed;

    if (argc > 2 || (argc == 2 && strcmp(argv[1], "setup") && strcmp(argv[1], "-setup")
		        && strcmp(argv[1], "--setup"))) {
        cerr << "Usage: either 'kam' for normal mode\n" "or 'kam setup' for setup mode" << endl;
	exit(ARG_ERROR);
    }
    if (argc == 2) prog_func.prog_mode = Prog_func::setup;

    configure_program(speed, port);

    if (Com::check_port(port) == -1) exit(COM_ERROR);  // check that port not in use

    Pipe_fifo receive_pipe(Pipe_fifo::non_block);
    Pipe_fifo send_pipe(Pipe_fifo::non_block);

    pid_t child_pid = fork();
    if (child_pid) {
        parent_process(child_pid, receive_pipe, send_pipe);
	delete[] prog_func.filedir;
	delete[] prog_func.myCall;
	delete[] prog_func.homedir;
    }
    else child_process(speed, port, receive_pipe, send_pipe);

    return 0;
}

//////////////////////////// function definitions //////////////////////////


/////////////////// parent process ////////////////////

void parent_process(pid_t child_pid, Pipe_fifo& receive_pipe, Pipe_fifo& send_pipe) {

    char combuffer[PIPE_BUF];

    Transmit_buffer tr_buffer(WORDWRAP);

    Tnc tnc(tr_buffer);
    Tnc_setup tnc_setup(tr_buffer);
    Tnc_base* tnc_base_p;
    if (prog_func.prog_mode == Prog_func::normal) tnc_base_p = &tnc;
    else tnc_base_p = &tnc_setup;

    MainScreen mainscreen(curses_init, tnc);
    ReceiveWin receivewin(tnc);
    SendWin sendwin;
    WinList screenform;
    screenform.add(&mainscreen);
    screenform.add(&receivewin);
    screenform.add(&sendwin);

    if (prog_func.GPL_flag == FALSE) {
        gpl_query();
	refresh_winlist(screenform);
    }	

    Keyboard_input_with_data* keyboard_p;
    if (prog_func.prog_mode == Prog_func::normal) {
        keyboard_p = new Keyboard_controller(mainscreen, sendwin, receivewin,
				         screenform, send_pipe, tr_buffer, tnc);
    }
    else keyboard_p = new Keyboard_setup_mode(mainscreen, sendwin, receivewin,
				         screenform, send_pipe, tr_buffer, tnc);
    if (!keyboard_p) {
        cerr << "Memory allocation error in main()" << endl;
	exit(MEM_ERROR);
    }

    tnc.set_wins(&mainscreen, &receivewin);
    tnc_setup.set_wins(&mainscreen, &receivewin);
    
    int letters_to_send_flag = FALSE;
    int flush_chars = FALSE;

// set up our selcall if configure_program() has not found it
    if (*(prog_func.mySelCall) == 0) {
        if (tnc.is_validcall(prog_func.myCall)) tnc.make_selcall(prog_func.myCall, prog_func.mySelCall);
	else {
	    receivewin.write("You have a non-standard callsign, which means I cannot work out your "
			     "Amtor Selcall\nPlease enter it manually via F5, or better, amend the "
			     "Kam\nconfiguration file (kamrc) to state the SELCALL: and restart\n");
	    receivewin.winrefresh();
	}
    }

// now make the pipes unidirectional

    receive_pipe.make_readonly();
    send_pipe.make_writeonly();

    timeval timeout;

    fd_set readfds_settings;
    FD_ZERO(&readfds_settings);
    FD_SET(0, &readfds_settings);
    receive_pipe.set_read_fds(&readfds_settings);

    fd_set readfds;

// now make sure the callsign and selcall in the Kam match the ones in this program
    if (prog_func.prog_mode == Prog_func::normal) {
        char* commandmessage;
	int length = strlen(prog_func.myCall) + 8;
	if (length < selCall_SIZE + 10) length = selCall_SIZE + 10;
	if ((commandmessage = new char[length]) == 0) {
	    cerr << "Memory allocation error in main()" << endl;
	    exit(MEM_ERROR);
	}
	strcpy(commandmessage, "MYCALL ");
	strcat(commandmessage, prog_func.myCall);
	Keyboard_controller::send_kamcommand(commandmessage);
	usleep(100000);
	strcpy(commandmessage, "MYSELCAL ");
	strcat(commandmessage, prog_func.mySelCall);
	Keyboard_controller::send_kamcommand(commandmessage);
	delete[] commandmessage;
    }

// enter freebytes information
    tnc_base_p->find_freebytes();

// everything is set up
// now enter the main program loop

    while (!prog_func.exitflag) {  // exit if exitflag set

// set up to use select() on stdin and receive_pipe on each iteration through the loop

	timeout.tv_sec = 0;
	timeout.tv_usec = 100000;  // timeout so that any file to be sent and to be piped from
	                           // tr_buffer is made available and clock display updated
	readfds = readfds_settings;

        int result;

	if (select(FD_SETSIZE, &readfds, 0, 0, &timeout)) {
	//if (select(FD_SETSIZE, &readfds, 0, 0, 0)) {  // useful for testing, but commented out
	                                                // for normal compile

// deal with any input on stdin or with any file being sent
	
	    if (FD_ISSET(0, &readfds) || keyboard_p->get_sendfile_flag() != Keyboard_input_with_data::not_sending) {
	        ((Keyboard_input*)keyboard_p->get_keyboard_ptr())->action();
	    }

// extract and process anything in receive_pipe

	    if (receive_pipe.isset_read_fds(&readfds)) {  // something in receive_pipe
	        usleep(10000); // as the com port only runs at 9600 baud, we can let a few characters
	                       // accumulate in receive_pipe and be even more kind to system resources
		int index;
		while ((result = receive_pipe.read(combuffer, PIPE_BUF)) > 0) {
		    for (index = 0; index < result; index++) {
		        tnc_base_p->process_received_byte(combuffer[index]);
		    }
		}
	    }	
	}

// check if anything needs to be sent out of tr_buffer to send_pipe
// (this includes a check whether there is a command in tr_buffer (the
// only command normally placed in the buffer is the receive command 'E':
// others are placed directly in send_pipe)

	int letter = tr_buffer.view_letter();
	if (letter == '\n' || letter == CMDinbuffer) flush_chars = TRUE;

	if (keyboard_p->get_sendfile_flag() != Keyboard_input_with_data::not_sending
	      || flush_chars == TRUE
	      || (tnc.tnc_func.active_port
	          && tnc.tnc_func.hfmode != Tnc_func::packet
	          && ((prog_func.send_mode == Prog_func::word
	               && letter == ' ') || 
		      (prog_func.send_mode == Prog_func::guard 
		       && tr_buffer.letters_used() > GUARD_QUEUE
		       && letter == ' ')))
	      || (prog_func.prog_mode == Prog_func::setup
		  && letter != -1)) {
	    letters_to_send_flag = TRUE;
	}

// send anything to be sent

	if (letters_to_send_flag && tnc_base_p->is_freebytes()) {
	    tnc_base_p->send_to_tnc(send_pipe, flush_chars);
	    if (!tnc.get_unsent_frame_flag()) {
	        letters_to_send_flag = FALSE;
		flush_chars = FALSE;
	    }
	}

// get Kam information, see if clock display is to be updated and check free bytes in Kam buffer
	
	tnc_base_p->get_info();
	static count = 20;
	if (count == 20) {
	    mainscreen.display_time();
	    tnc_base_p->find_freebytes();
	    count = 0;
	}
	else count++;
    }

// prog_func.exitflag has been set - clean up and then quit

    //Keyboard_controller::send_specialcommand('Q');       // quit HOST mode

// kill child process
    kill(child_pid, SIGQUIT);

    receivewin.write("\nClosing down ... \n");
    receivewin.winrefresh();

// make sure that if any temporary file is deleted
    char* tempfile = new char[strlen(prog_func.filedir) + 11];
    if (!tempfile) {
        cerr << "Memory allocation error in parent_process()" << endl;
	exit(MEM_ERROR);
    }
    strcpy(tempfile, prog_func.filedir);
    strcat(tempfile, "/temp_file");
    unlink(tempfile);
    delete[] tempfile;
    delete keyboard_p;
}

/////////////////////// child process ////////////////////////////

void child_process(int speed, int port, Pipe_fifo& receive_pipe, Pipe_fifo& send_pipe) {

// set up the signal that will tell this process to terminate
    child_exit_flag = FALSE;
    signal(SIGQUIT, child_exit_handler);

// set up variables and objects for this process
    char combuffer[PIPE_BUF];
    Com com_obj(speed, port);

// make the pipes unidirectional

    receive_pipe.make_writeonly();
    send_pipe.make_readonly();

    if (prog_func.prog_mode == Prog_func::normal) {
// make sure the Kam is in host mode

        com_obj.send(3);           // make sure we are in command mode if not already in host mode
	usleep(100000);
	com_obj.send('\r');        // clear out any debris in the file buffer
	usleep(100000);
	com_obj.send((unsigned char*)"INTFACE HOST\r");        // put the Kam in host mode
	sleep(1);
	com_obj.send((unsigned char*)"RESET\r");
	usleep(100000);
    }
    else {  // we must be trying to set up the Kam
// make sure we are not in host mode
	usleep(100000);
        com_obj.send(FEND);
	com_obj.send('Q');
        com_obj.send(FEND);
    }

    char letter;
    int result1;
    int result2;
    int index;
    while (!child_exit_flag) {
        // receive input and insert into pipe

        for (index = 0; index < PIPE_BUF && (result1 = com_obj.receive()) != EOF; index++) {
	    combuffer[index] = (char)result1;
	}
	if (index > 0) do {
	    result2 = receive_pipe.write(combuffer, index);
	} while (!result2);  // keep going until pipe will accept input

	// now send anything waiting in pipe to be sent

	if ((result1 = send_pipe.read(combuffer, PIPE_BUF)) > 0) {
	    for(index = 0; index < result1; index++) {
	        letter = combuffer[index];
		do {
		    result2 = com_obj.send((uchar)letter);
		} while (result2 == EOF);  // keep going until com port will accept output
	    }
	}
    }
}

///////////////////////////////////////////////////////////


void configure_program(int& speed, int& port) {
    char* home = getenv("HOME");
    int found_rcfile = FALSE;
    if (home) {
        prog_func.homedir = new char[strlen(home) + 1];
	if (!prog_func.homedir) {
	    cerr << "Memory allocation error in configure_program()" << endl;
	    exit(MEM_ERROR);
	}
	strcpy(prog_func.homedir, home);

        char* home_rcfile = new char[strlen(home) + strlen(RC_FILE) + 2];
	if (!home_rcfile) {
	    cerr << "Memory allocation error in configure_program()" << endl;
	    exit(MEM_ERROR);
	}
	strcpy(home_rcfile, home);
	strcat(home_rcfile, "/");
	strcat(home_rcfile, RC_FILE);

	filein.open(home_rcfile, ios::in | ios::nocreate);
	if (filein) found_rcfile = TRUE;
	else filein.clear();
	delete[] home_rcfile;
    }
    if (found_rcfile == FALSE) {
        char* system_rcfile = new char[strlen(RC_FILE) + 6];
	if (!system_rcfile) {
	    cerr << "Memory allocation error in configure_program()" << endl;
	    exit(MEM_ERROR);
	}
	strcpy(system_rcfile, "/etc/");
	strcat(system_rcfile, RC_FILE);
	filein.open(system_rcfile, ios::in | ios::nocreate);
	if (filein) found_rcfile = TRUE;
	else filein.clear();
	delete[] system_rcfile;  // no need to delete the storage now
    }
    if (found_rcfile == FALSE) {
        cerr << "Can't find or open file /etc/" RC_FILE << endl;
	cerr << "or file $HOME/." RC_FILE << endl;
        exit(FILEOPEN_ERROR);
    }

// check to see if GPL terms have been accepted

    char* gpl_file = new char[strlen(prog_func.homedir) + 9];
    if (!gpl_file) {
        cerr << "Memory allocation error in configure_program()" << endl;
	exit(MEM_ERROR);
    }
    strcpy(gpl_file, prog_func.homedir);
    strcat(gpl_file, "/");
    strcat(gpl_file, ".kamgpl");
    ifstream gplfile(gpl_file, ios::in | ios::nocreate);
    if (gplfile) {
        prog_func.GPL_flag = TRUE;
	gplfile.close();
    }

// now extract settings from file

    char buffer[257];
    char* buffer_ptr;
    int count;
    int call_found_flag = FALSE;
    int port_found_flag = FALSE;
    int speed_found_flag = FALSE;
    int filedir_found_flag = FALSE;
    while(filein.getline(buffer, 256)) {
	if ((count = filein.gcount()) > 1 && *buffer != '#') {  // # at start of line is a comment
	    buffer[--count] = 0;  // null terminate the string

	    if ((buffer_ptr = strstr(buffer, "CALL:")) != 0) {  // found CALL:
	        buffer_ptr += 5;        // advance buffer_ptr to the letter after CALL:

		while (*buffer_ptr == ' ') buffer_ptr++;  // get rid of leading white space

		char* temp_ptr = buffer_ptr; // get rid of any trailing white space or comment
		while(*temp_ptr != 0 && *temp_ptr != ' ' && *temp_ptr != '#') temp_ptr++;
		if (*temp_ptr == ' ' || *temp_ptr == '#') *temp_ptr = 0;

		for(temp_ptr = buffer_ptr; *temp_ptr != 0; temp_ptr++) {  // convert to upper case
		    *temp_ptr = toupper(*temp_ptr);
		}

		int length;
		if ((length = strlen(buffer_ptr)) != 0) {
		    prog_func.myCall = new char[length + 1];
		    if (!prog_func.myCall) {
		        cerr << "Memory allocation error in configure_program(int&, int&)" << endl;
			exit(MEM_ERROR);
		    }
		    strcpy(prog_func.myCall, buffer_ptr);
		    call_found_flag = TRUE;
		}
	    }
	    
	    else if ((buffer_ptr = strstr(buffer, "SELCALL:")) != 0) { // found SELCALL:
	        buffer_ptr += 8;        // advance buffer_ptr to the letter after SELCALL:

		while (*buffer_ptr == ' ') buffer_ptr++;  // get rid of leading white space

		char* temp_ptr = buffer_ptr; // get rid of any trailing white space or comment
		while(*temp_ptr != 0 && *temp_ptr != ' ' && *temp_ptr != '#') temp_ptr++;
		if (*temp_ptr == ' ' || *temp_ptr == '#') *temp_ptr = 0;

		for(temp_ptr = buffer_ptr; *temp_ptr != 0; temp_ptr++) {  // convert to upper case
		    *temp_ptr = toupper(*temp_ptr);
		}

		int length;
		if ((length = strlen(buffer_ptr)) != 0 && length <= selCall_SIZE) {
		    strcpy(prog_func.mySelCall, buffer_ptr);
		}
	    }

	    else if ((buffer_ptr = strstr(buffer, "PORT:")) != 0) { // found PORT:
	        buffer_ptr += 5;        // advance buffer_ptr to the letter after PORT:

		while (*buffer_ptr == ' ') buffer_ptr++;  // get rid of leading white space

		char* temp_ptr = buffer_ptr; // get rid of any trailing white space or comment
		while(*temp_ptr != 0 && *temp_ptr != ' ' && *temp_ptr != '#') temp_ptr++;
		if (*temp_ptr == ' ' || *temp_ptr == '#') *temp_ptr = 0;

		int length;
		if ((length = strlen(buffer_ptr)) != 0) {
		    port = atoi(buffer_ptr);
		    if (port < 1 || port > 4) {
		        filein.close();
			cerr << "Serial port 1 to 4 must be specified in tnc.conf" << endl;
			exit(COM_ERROR);
		    }
		    port_found_flag = TRUE;
		}
	    }

	    else if ((buffer_ptr = strstr(buffer, "FILEDIR:")) != 0) {  // found FILEDIR:
	        buffer_ptr += 8;        // advance buffer_ptr to the letter after FILEDIR:

		while (*buffer_ptr == ' ') buffer_ptr++;  // get rid of leading white space

		char* temp_ptr = buffer_ptr; // get rid of any trailing white space or comment
		while(*temp_ptr != 0 && *temp_ptr != ' ' && *temp_ptr != '#') temp_ptr++;
		if (*temp_ptr == ' ' || *temp_ptr == '#') *temp_ptr = 0;

		int length;
		if ((length = strlen(buffer_ptr)) != 0) {
		    prog_func.filedir = new char[length + 1];
		    if (!prog_func.filedir) {
		        cerr << "Memory allocation error in configure_program(int&, int&, char*&)" << endl;
			exit(MEM_ERROR);
		    }
		    strcpy(prog_func.filedir, buffer_ptr);
		    filedir_found_flag = TRUE;
		}
	    }

	    else if ((buffer_ptr = strstr(buffer, "SPEED:")) != 0) { // found SPEED:
	        buffer_ptr += 6;        // advance buffer_ptr to the letter after SPEED:

		while (*buffer_ptr == ' ') buffer_ptr++;  // get rid of leading white space

		char* temp_ptr = buffer_ptr; // get rid of any trailing white space or comment
		while(*temp_ptr != 0 && *temp_ptr != ' ' && *temp_ptr != '#') temp_ptr++;
		if (*temp_ptr == ' ' || *temp_ptr == '#') *temp_ptr = 0;

		speed = atoi(buffer_ptr);
		if (speed == 300 || speed == 600 || speed == 1200 ||
		    speed == 2400 || speed == 4800 || speed == 9600) {
		    speed_found_flag = TRUE;
		}
	    }

	    else if ((buffer_ptr = strstr(buffer, "NOPROMPT:")) != 0) { // found NOPROMPT:
	        buffer_ptr += 9;        // advance buffer_ptr to the letter after NOPROMPT:

		while (*buffer_ptr == ' ') buffer_ptr++;  // get rid of leading white space

		char* temp_ptr = buffer_ptr; // get rid of any trailing white space or comment
		while(*temp_ptr != 0 && *temp_ptr != ' ' && *temp_ptr != '#') temp_ptr++;
		if (*temp_ptr == ' ' || *temp_ptr == '#') *temp_ptr = 0;

		for(temp_ptr = buffer_ptr; *temp_ptr != 0; temp_ptr++) {  // convert to upper case
		    *temp_ptr = toupper(*temp_ptr);
		}
		if (!strcmp("TRUE", buffer_ptr)) prog_func.noprompt_flag = TRUE;
	    }
	}
    }

    if (!call_found_flag) {
        filein.close();
        cerr << "Your callsign must be specified in kamdir" << endl;
	cerr << "Include a 'CALL: xxxx' statement in that file, where xxxx "
	            "is your callsign" << endl;
    }
    if (!port_found_flag) {
        filein.close();
        cerr << "Serial port 1 to 4 must be specified in kamdir" << endl;
	cerr << "Include a 'PORT: x' statement in that file, where x is the "
	            "com port used with the TNC" << endl;
    }
    if (!filedir_found_flag) {
        filein.close();
	cerr << "The directory in which files may be found must be specified in kamrc" << endl;
	cerr << "Include a 'FILEDIR: xxxxxxxx' statement in that file, where xxxxxxxx "
	            "is the full directory path" << endl;
    }
    if (!speed_found_flag) speed = 9600;
    filein.close();
    if (!call_found_flag || !port_found_flag || !filedir_found_flag) exit(CONFIG_ERROR);
}

void send_parms(WinList& screenform) {

    char* parm_file = new char[strlen(prog_func.filedir) + 11];
    if (!parm_file) {
        cerr << "Memory allocation error in send_parms(WinList&)" << endl;
	exit(MEM_ERROR);
    }
    strcpy(parm_file, prog_func.filedir);
    strcat(parm_file, "/tnc.parms");
    
    filein.open(parm_file, ios::in | ios::nocreate);
    delete[] parm_file;
    if (!filein) {
        PopupWin* popupwin_p = new PopupWin;
	if (!popupwin_p) {
	    cerr << "Memory allocation error in send_parms(WinList&)" << endl;
	    exit(MEM_ERROR);
	}
        popupwin_p->show();
	popupwin_p->write("File tnc.parms", 1, 6);
	popupwin_p->write("cannot be found/opened", 2, 6);
	popupwin_p->write("Check that FILEDIR: in kamrc is correct", 3, 6);
	sleep(3);
	popupwin_p->unshow();
	delete popupwin_p;
	refresh_winlist(screenform);
	filein.clear();
    }
    else {
        char buffer[258];
	int count;
	while(filein.getline(buffer, 256)) {
	    if ((count = filein.gcount()) > 1 && *buffer != '#') {  // # at start of line is a comment
	        buffer[--count] = 0;   // null terminate the string
		
		char* buffer_ptr = buffer; // get rid of any trailing comment
		while(*buffer_ptr != 0 && *buffer_ptr != '#') buffer_ptr++;
		if (*buffer_ptr == '#') {
		    *buffer_ptr = 0;
		}
		Keyboard_controller::send_kamcommand(buffer);
		usleep(100000);
	    }
	}
	filein.close();
    }
    char* commandmessage;
    int length = strlen(prog_func.myCall) + 8;
    if (length < selCall_SIZE + 10) length = selCall_SIZE + 10;
    if ((commandmessage = new char[length]) == 0) {
        cerr << "Memory allocation error in sendparms(WinList&)" << endl;
	exit(MEM_ERROR);
    }
    strcpy(commandmessage, "MYCALL ");
    strcat(commandmessage, prog_func.myCall);
    Keyboard_controller::send_kamcommand(commandmessage);
    usleep(100000);
    strcpy(commandmessage, "MYSELCAL ");
    strcat(commandmessage, prog_func.mySelCall);
    Keyboard_controller::send_kamcommand(commandmessage);
    delete[] commandmessage;
}


void gpl_query(void) {
    CopyrightScreen copyright_screen;
    Gpl_query gpl_query(copyright_screen);

    fd_set readfds_settings, readfds;
    FD_ZERO(&readfds_settings);
    FD_SET(0, &readfds_settings);

    while(!prog_func.GPL_flag) {
    	readfds = readfds_settings;
    	select(FD_SETSIZE, &readfds, 0, 0, 0);
	gpl_query.Gpl_query::action();
    }
}

void child_exit_handler(int signal) {
    child_exit_flag = signal;
}
